/* This file is part of Cloudy and is copyright (C)1978-2015 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */
#include "cddefines.h"
#include "vectorize.h"
#include "vectorize_math.h"

//
// Written by Peter A.M. van Hoof, Royal Observatory of Belgium, Brussels
//
// this file contains vectorized versions of the single and double variants of the exp()
// function. They are vectorized using AVX instructions, but also make use of AVX2, FMA,
// and AVX512 instructions when available. The basic algorithms for calculating the exp()
// functions were slightly modified from versions taken from the Cephes library written by
// Stephen L. Moshier, available at: http://www.netlib.org/cephes/
//
// This work was also inspired by similar routines written by Giovanni Garberoglio which
// are available at: http://software-lisc.fbk.eu/avx_mathfun/
//

#ifdef __AVX__

VECD_CONST(exp_max_arg,+709.782712894);
VECD_CONST(exp_min_arg,-709.089565712);

VECD_CONST(log2e,1./log(2.));

#ifdef __AVX512DQ__
VECL_CONST(off,0x3ff);
#else
VECI_CONST(off,0x3ff);
#endif

VECD_CONST(exp_c1,-6.93145751953125e-1);
VECD_CONST(exp_c2,-1.42860682030941723212e-6);

VECD_CONST(exp_p0,1.26177193074810590878e-4);
VECD_CONST(exp_p1,3.02994407707441961300e-2);
VECD_CONST(exp_p2,1.00000000000000000000e+0);
VECD_CONST(exp_q0,3.00198505138664455042e-6);
VECD_CONST(exp_q1,2.52448340349684104192e-3);
VECD_CONST(exp_q2,2.27265548208155028766e-1);
VECD_CONST(exp_q3,2.000000000000000000000e0);

VECD_CONST(two,2.);
VECD_CONST(one,1.);

VECD_CONST(ln_ten,log(10.));

#ifdef __AVX512F__

inline v8df v1expd(v8df x)
{
	x = _mm512_min_pd(x, exp_max_arg);
	x = _mm512_max_pd(x, exp_min_arg);
	v8df px = _mm512_mul_pd(x, log2e);
	px = _mm512_floor_pd(px);
#ifdef __AVX512DQ__
	v8di n = _mm512_cvtpd_epi64(px);
	n = _mm512_add_epi64(n, off);
#else
	v8si m = _mm512_cvtpd_epi32(px);
	m = _mm256_add_epi32(m, off);
	v8di n = _mm512_cvtepi32_epi64(m);
#endif
	x = _mm512_fmadd_pd(px, exp_c1, x);
	x = _mm512_fmadd_pd(px, exp_c2, x);
	v8df xx = _mm512_mul_pd(x, x);
	px = _mm512_fmadd_pd(exp_p0, xx, exp_p1);
	px = _mm512_fmadd_pd(px, xx, exp_p2);
	v8df qx = _mm512_fmadd_pd(exp_q0, xx, exp_q1);
	qx = _mm512_fmadd_pd(qx, xx, exp_q2);
	qx = _mm512_fmadd_pd(qx, xx, exp_q3);
	px = _mm512_mul_pd(x, px);
	xx = _mm512_sub_pd(qx, px);
	x = _mm512_div_pd(px, xx);
	x = _mm512_fmadd_pd(two, x, one);
	n = _mm512_slli_epi64(n, 52);
	v8df p2n = _mm512_castsi512_pd(n);
	return _mm512_mul_pd(x, p2n);
}

inline v8df v1exp10d(v8df x)
{
	x = _mm512_mul_pd(x, ln_ten);
	return v1expd(x);
}

#else

inline v4df v1expd(v4df x)
{
	x = _mm256_min_pd(x, exp_max_arg);
	x = _mm256_max_pd(x, exp_min_arg);
	v4df px = _mm256_mul_pd(x, log2e);
	px = _mm256_floor_pd(px);
	v4si m = _mm256_cvtpd_epi32(px);
	m = _mm_add_epi32(m, off);
#ifdef __FMA__
	x = _mm256_fmadd_pd(px, exp_c1, x);
	x = _mm256_fmadd_pd(px, exp_c2, x);
#else
	v4df y = _mm256_mul_pd(px, exp_c1);
	x = _mm256_add_pd(x, y);
	y = _mm256_mul_pd(px, exp_c2);
	x = _mm256_add_pd(x, y);
#endif
	v4df xx = _mm256_mul_pd(x, x);
#ifdef __FMA__
	px = _mm256_fmadd_pd(exp_p0, xx, exp_p1);
	px = _mm256_fmadd_pd(px, xx, exp_p2);
	v4df qx = _mm256_fmadd_pd(exp_q0, xx, exp_q1);
	qx = _mm256_fmadd_pd(qx, xx, exp_q2);
	qx = _mm256_fmadd_pd(qx, xx, exp_q3);
#else
	px = _mm256_mul_pd(exp_p0, xx);
	px = _mm256_add_pd(px, exp_p1);
	px = _mm256_mul_pd(px, xx);
	px = _mm256_add_pd(px, exp_p2);
	v4df qx = _mm256_mul_pd(exp_q0, xx);
	qx = _mm256_add_pd(qx, exp_q1);
	qx = _mm256_mul_pd(qx, xx);
	qx = _mm256_add_pd(qx, exp_q2);
	qx = _mm256_mul_pd(qx, xx);
	qx = _mm256_add_pd(qx, exp_q3);
#endif
	px = _mm256_mul_pd(x, px);
	xx = _mm256_sub_pd(qx, px);
	x = _mm256_div_pd(px, xx);
#ifdef __FMA__
	x = _mm256_fmadd_pd(two, x, one);
#else
	x = _mm256_mul_pd(x, two);
	x = _mm256_add_pd(x, one);
#endif
#ifdef __AVX2__
	v4di n = _mm256_cvtepi32_epi64(m);
	n = _mm256_slli_epi64(n, 52);
#else
	m = _mm_slli_epi32(m, 20);
	v4si z = _mm_set1_epi32(0);
	v4si nl = _mm_unpacklo_epi32(z, m);
	v8si n = _mm256_set1_epi32(0);
	n = _mm256_insertf128_si256(n, nl, 0);
	v4si nu = _mm_unpackhi_epi32(z, m);
	n = _mm256_insertf128_si256(n, nu, 1);
#endif
	v4df p2n = _mm256_castsi256_pd(n);
	return _mm256_mul_pd(x, p2n);
}

inline v4df v1exp10d(v4df x)
{
	x = _mm256_mul_pd(x, ln_ten);
	return v1expd(x);
}

#endif // __AVX512F__

VECF_CONST(expf_max_arg,+88.722840f);
VECF_CONST(expf_min_arg,-88.029691f);

VECF_CONST(log2ef,1.f/logf(2.f));

#if defined(__AVX2__) || defined(__AVX512F__)
VECII_CONST(offf,0x7f);
#else
VECI_CONST(offf,0x7f);
#endif

VECF_CONST(expf_c1,-0.693359375f);
VECF_CONST(expf_c2,2.12194440e-4f);

VECF_CONST(expf_p0,1.9875691500e-4f);
VECF_CONST(expf_p1,1.3981999507e-3f);
VECF_CONST(expf_p2,8.3334519073e-3f);
VECF_CONST(expf_p3,4.1665795894e-2f);
VECF_CONST(expf_p4,1.6666665459e-1f);
VECF_CONST(expf_p5,5.0000001201e-1f);

VECF_CONST(onef,1.f);

VECF_CONST(ln_tenf,logf(10.f));

#ifdef __AVX512F__

inline v16sf v1expf(v16sf x)
{
	x = _mm512_min_ps(x, expf_max_arg);
	x = _mm512_max_ps(x, expf_min_arg);
	v16sf px = _mm512_mul_ps(x, log2ef);
	px = _mm512_floor_ps(px);
	v16si n = _mm512_cvtps_epi32(px);
	n = _mm512_add_epi32(n, offf);
	x = _mm512_fmadd_ps(px, expf_c1, x);
	x = _mm512_fmadd_ps(px, expf_c2, x);
	v16sf xx = _mm512_mul_ps(x, x);
	px = _mm512_fmadd_ps(expf_p0, x, expf_p1);
	px = _mm512_fmadd_ps(px, x, expf_p2);
	px = _mm512_fmadd_ps(px, x, expf_p3);
	px = _mm512_fmadd_ps(px, x, expf_p4);
	px = _mm512_fmadd_ps(px, x, expf_p5);
	px = _mm512_mul_ps(px, xx);
	px = _mm512_add_ps(px, x);
	px = _mm512_add_ps(px, onef);
	n = _mm512_slli_epi32(n, 23);
	v16sf p2n = _mm512_castsi512_ps(n);
	return _mm512_mul_ps(px, p2n);
}

inline v16sf v1exp10f(v16sf x)
{
	x = _mm512_mul_ps(x, ln_tenf);
	return v1expf(x);
}

#else

inline v8sf v1expf(v8sf x)
{
	x = _mm256_min_ps(x, expf_max_arg);
	x = _mm256_max_ps(x, expf_min_arg);
	v8sf px = _mm256_mul_ps(x, log2ef);
	px = _mm256_floor_ps(px);
	v8si n = _mm256_cvtps_epi32(px);
#ifdef __AVX2__
	n = _mm256_add_epi32(n, offf);
#else
	v4si nl = _mm256_extractf128_si256(n, 0);
	nl = _mm_add_epi32(nl, offf);
	v4si nu = _mm256_extractf128_si256(n, 1);
	nu = _mm_add_epi32(nu, offf);
#endif
#ifdef __FMA__
	x = _mm256_fmadd_ps(px, expf_c1, x);
	x = _mm256_fmadd_ps(px, expf_c2, x);
#else
	v8sf y = _mm256_mul_ps(px, expf_c1);
	x = _mm256_add_ps(x, y);
	y = _mm256_mul_ps(px, expf_c2);
	x = _mm256_add_ps(x, y);
#endif
	v8sf xx = _mm256_mul_ps(x, x);
#ifdef __FMA__
	px = _mm256_fmadd_ps(expf_p0, x, expf_p1);
	px = _mm256_fmadd_ps(px, x, expf_p2);
	px = _mm256_fmadd_ps(px, x, expf_p3);
	px = _mm256_fmadd_ps(px, x, expf_p4);
	px = _mm256_fmadd_ps(px, x, expf_p5);
#else
	px = _mm256_mul_ps(expf_p0, x);
	px = _mm256_add_ps(px, expf_p1);
	px = _mm256_mul_ps(px, x);
	px = _mm256_add_ps(px, expf_p2);
	px = _mm256_mul_ps(px, x);
	px = _mm256_add_ps(px, expf_p3);
	px = _mm256_mul_ps(px, x);
	px = _mm256_add_ps(px, expf_p4);
	px = _mm256_mul_ps(px, x);
	px = _mm256_add_ps(px, expf_p5);
#endif
	px = _mm256_mul_ps(px, xx);
	px = _mm256_add_ps(px, x);
	px = _mm256_add_ps(px, onef);
#ifdef __AVX2__
	n = _mm256_slli_epi32(n, 23);
#else
	nl = _mm_slli_epi32(nl, 23);
	nu = _mm_slli_epi32(nu, 23);
	n = _mm256_insertf128_si256(n, nl, 0);
	n = _mm256_insertf128_si256(n, nu, 1);
#endif
	v8sf p2n = _mm256_castsi256_ps(n);
	return _mm256_mul_ps(px, p2n);
}

inline v8sf v1exp10f(v8sf x)
{
	x = _mm256_mul_ps(x, ln_tenf);
	return v1expf(x);
}

#endif // __AVX512F__

#else

// stub routines, should never be called
inline int v1expd(int) { return 0; }
inline int v1exp10d(int) { return 0; }
inline int v1expf(int) { return 0; }
inline int v1exp10f(int) { return 0; }

#endif // __AVX__

// wrapper routines to give math functions C++ linkage
// this prevents warnings from the Oracle Studio compiler
inline double wr_expd(double x)
{
	return exp(x);
}

inline sys_float wr_expf(sys_float x)
{
	return expf(x);
}

// wrapper routines needed to resolve overloaded functions
// these prevent problems with the Oracle Studio compiler
// it cannot handle overloaded functions as a template argument
inline double wr_exp10d(double x)
{
	return exp10(x);
}

inline sys_float wr_exp10f(sys_float x)
{
	return exp10(x);
}


void vexp(const double x[], double y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vexp()" );

	vecfun( x, y, nlo, nhi, wr_expd, v1expd );
}

void vexp10(const double x[], double y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vexp10()" );

	vecfun( x, y, nlo, nhi, wr_exp10d, v1exp10d );
}

void vexp(const sys_float x[], sys_float y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vexp()" );

	vecfun( x, y, nlo, nhi, wr_expf, v1expf );
}

void vexp10(const sys_float x[], sys_float y[], long nlo, long nhi)
{
	DEBUG_ENTRY( "vexp10()" );

	vecfun( x, y, nlo, nhi, wr_exp10f, v1exp10f );
}

void vexp(double *y, double x0, double x1, double x2, double x3)
{
	V1FUN_PD_4(exp, 0.);
}

void vexp10(double *y, double x0, double x1, double x2, double x3)
{
	V1FUN_PD_4(exp10, 0.);
}

void vexp(double *y, double x0, double x1, double x2, double x3, double x4, double x5, double x6, double x7)
{
	V1FUN_PD_8(exp, 0.);
}

void vexp10(double *y, double x0, double x1, double x2, double x3, double x4, double x5, double x6, double x7)
{
	V1FUN_PD_8(exp10, 0.);
}

void vexp(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3)
{
	V1FUN_PS_4(exp, 0.f);
}

void vexp10(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3)
{
	V1FUN_PS_4(exp10, 0.f);
}

void vexp(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5, 
	  sys_float x6, sys_float x7)
{
	V1FUN_PS_8(exp, 0.f);
}

void vexp10(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5,
	    sys_float x6, sys_float x7)
{
	V1FUN_PS_8(exp10, 0.f);
}

void vexp(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5, 
	  sys_float x6, sys_float x7, sys_float x8, sys_float x9, sys_float x10, sys_float x11, sys_float x12,
	  sys_float x13, sys_float x14, sys_float x15)
{
	V1FUN_PS_16(exp, 0.f);
}

void vexp10(sys_float *y, sys_float x0, sys_float x1, sys_float x2, sys_float x3, sys_float x4, sys_float x5, 
	    sys_float x6, sys_float x7, sys_float x8, sys_float x9, sys_float x10, sys_float x11, sys_float x12,
	    sys_float x13, sys_float x14, sys_float x15)
{
	V1FUN_PS_16(exp10, 0.f);
}
