/* This file is part of Cloudy and is copyright (C)1978-2015 by Gary J. Ferland and
 * others.  For conditions of distribution and use see copyright notice in license.txt */

#ifndef VECTORIZE_MATH_H
#define VECTORIZE_MATH_H

// uint64 is required for AVX support -- uint64 should be present on all modern systems
#ifndef HAVE_UINT64
#undef __AVX__
#endif

#ifdef __AVX__

// unconditionally disable avx512f optimizations since these have not been tested yet
// remove the following line once the testing is complete...
#undef __AVX512F__

// use shorthand for packed types
typedef __m128d v2df;
typedef __m128  v4sf;
typedef __m128i v2di;
typedef __m128i v4si;

typedef __m256d v4df;
typedef __m256  v8sf;
typedef __m256i v4di;
typedef __m256i v8si;

#ifdef __AVX512F__

typedef __m512d v8df;
typedef __m512  v16sf;
typedef __m512i v8di;
typedef __m512i v16si;

// some macros for defining vector constants

#define VECD_CONST(name,x) \
ALIGNED(64) static const double __avx_##name[8] = {x,x,x,x,x,x,x,x}; \
static const v8df& name = *reinterpret_cast<const v8df*>(__avx_##name)

#define VECF_CONST(name,x) \
ALIGNED(64) static const sys_float __avx_##name[16] = {x,x,x,x,x,x,x,x,x,x,x,x,x,x,x,x}; \
static const v16sf& name = *reinterpret_cast<const v16sf*>(__avx_##name)

#define VECI_CONST(name,x) \
ALIGNED(32) static const uint32 __avx_##name[8] = {x,x,x,x,x,x,x,x}; \
static const v8si& name = *reinterpret_cast<const v8si*>(__avx_##name)

#define VECII_CONST(name,x) \
ALIGNED(64) static const uint32 __avx_##name[16] = {x,x,x,x,x,x,x,x,x,x,x,x,x,x,x,x}; \
static const v16si& name = *reinterpret_cast<const v16si*>(__avx_##name)

#define VECL_CONST(name,x) \
ALIGNED(64) static const uint64 __avx_##name[8] = {x,x,x,x,x,x,x,x}; \
static const v8di& name = *reinterpret_cast<const v8di*>(__avx_##name)

union vpack
{
	ALIGNED(64) v8df pd;
	ALIGNED(64) v16sf ps;
	ALIGNED(64) double d[8];
	ALIGNED(64) sys_float x[16];
};

#else // __AVX512F__

#define VECD_CONST(name,x) \
ALIGNED(32) static const double __avx_##name[4] = {x,x,x,x}; \
static const v4df& name = *reinterpret_cast<const v4df*>(__avx_##name)

#define VECF_CONST(name,x) \
ALIGNED(32) static const sys_float __avx_##name[8] = {x,x,x,x,x,x,x,x}; \
static const v8sf& name = *reinterpret_cast<const v8sf*>(__avx_##name)

#define VECI_CONST(name,x) \
ALIGNED(16) static const uint32 __avx_##name[4] = {x,x,x,x}; \
static const v4si& name = *reinterpret_cast<const v4si*>(__avx_##name)

#define VECII_CONST(name,x) \
ALIGNED(32) static const uint32 __avx_##name[8] = {x,x,x,x,x,x,x,x}; \
static const v8si& name = *reinterpret_cast<const v8si*>(__avx_##name)

#define VECL_CONST(name,x) \
ALIGNED(32) static const uint64 __avx_##name[4] = {x,x,x,x}; \
static const v4di& name = *reinterpret_cast<const v4di*>(__avx_##name)

union vpack
{
	ALIGNED(32) v4df pd;
	ALIGNED(32) v8sf ps;
	ALIGNED(32) double d[4];
	ALIGNED(32) sys_float f[8];
};

#endif // __AVX512F__

template<class V>
inline void vecfun_partial(const sys_float x[], sys_float y[], long nlo, long nhi, long npack, V (*vecfun1)(V))
{
	vpack xx, yy;
	for( long i=0; i < npack; ++i )
		xx.f[i] = x[min(nlo+i,nhi-1)];
	yy.ps = vecfun1(xx.ps);
	for( long i=nlo; i < nhi; ++i )
		y[i] = yy.f[i-nlo];
}

template<class V>
inline void vecfun_partial(const double x[], double y[], long nlo, long nhi, long npack, V (*vecfun1)(V))
{
	vpack xx, yy;
	for( long i=0; i < npack; ++i )
		xx.d[i] = x[min(nlo+i,nhi-1)];
	yy.pd = vecfun1(xx.pd);
	for( long i=nlo; i < nhi; ++i )
		y[i] = yy.d[i-nlo];
}

// this is the generic wrapper routine for all vectorized math functions
// it makes sure that the vectorized function gets arguments that are properly aligned
// scalfun1 is the normal scalar routine working on a single argument
// vecfun1 is the vectorized routine working on a packed register
template<class T, class V>
inline void vecfun(const T x[], T y[], long nlo, long nhi, T (*)(T), V (*vecfun1)(V))
{
	if( nhi <= nlo )
		return;

	// determine number of elements of type T in a packed register
	long npack = CD_ALIGN/sizeof(T);
	if( nhi-nlo < npack )
	{
		vecfun_partial(x, y, nlo, nhi, npack, vecfun1);
		return;
	}
	long i, ilo = nlo;
	long ox = (reinterpret_cast<long>(x)/sizeof(T)+nlo)%npack;
	long oy = (reinterpret_cast<long>(y)/sizeof(T)+nlo)%npack;
	bool lgNonAligned = ( ox != oy );
	T *yl, *ylocal = NULL;
	if( lgNonAligned )
	{
		ylocal = new T[nhi+npack-1];
		long ol = (reinterpret_cast<long>(ylocal)/sizeof(T)+nlo)%npack;
		if( ox >= ol )
			yl = &ylocal[ox-ol];
		else
			yl = &ylocal[ox+npack-ol];
	}
	else
	{
		yl = y;
	}
	// the initial element is not aligned correctly ->
	// use scalar routine until we are correctly aligned
	if( ox > 0 )
	{
		vecfun_partial(x, yl, ilo, min(ilo+npack-ox,nhi), npack, vecfun1);
		ilo = min(ilo+npack-ox,nhi);
	}
	// use vectorized routine as long as there are at least npack evaluations left
	for( i=ilo; i < nhi-npack+1; i += npack )
	{
		const V& xx = *reinterpret_cast<const V*>(&x[i]);
		V& yy = *reinterpret_cast<V*>(&yl[i]);
		yy = vecfun1(xx);
	}
	ilo = i;
	// use scalar routine again for the remaining calls (if any)...
	vecfun_partial(x, yl, ilo, nhi, npack, vecfun1);
	if( lgNonAligned )
	{
		for( i=nlo; i < nhi; ++i )
			y[i] = yl[i];
		delete[] ylocal;
	}
}

#ifdef __AVX512F__
#define V1FUN_PD_4(FUN, V) \
	v8df xx = _mm512_set_pd( V, V, V, V, x3, x2, x1, x0 ); \
	v8df yy = v1##FUN##d(xx); \
	memcpy(y, &yy, 4*sizeof(double))
#else
#define V1FUN_PD_4(FUN, V) \
	v4df xx = _mm256_set_pd( x3, x2, x1, x0 ); \
	v4df yy = v1##FUN##d(xx); \
	memcpy(y, &yy, 4*sizeof(double))
#endif

#ifdef __AVX512F__
#define V1FUN_PD_8(FUN, V) \
	v8df xx = _mm512_set_pd( x7, x6, x5, x4, x3, x2, x1, x0 ); \
	v8df yy = v1##FUN##d(xx); \
	memcpy(y, &yy, 8*sizeof(double))
#else
#define V1FUN_PD_8(FUN, V) \
	v4df yy[2]; \
	v4df xx = _mm256_set_pd( x3, x2, x1, x0 ); \
	yy[0] = v1##FUN##d(xx); \
	xx = _mm256_set_pd( x7, x6, x5, x4 ); \
	yy[1] = v1##FUN##d(xx); \
	memcpy(y, &yy, 8*sizeof(double))
#endif

#ifdef __AVX512F__
#define V1FUN_PS_4(FUN, V) \
	v16sf xx = _mm512_set_ps( V, V, V, V, V, V, V, V, V, V, V, V, x3, x2, x1, x0 ); \
	v16sf yy = v1##FUN##f(xx); \
	memcpy(y, &yy, 4*sizeof(sys_float))
#else
#define V1FUN_PS_4(FUN, V) \
	v8sf xx = _mm256_set_ps( V, V, V, V, x3, x2, x1, x0 ); \
	v8sf yy = v1##FUN##f(xx); \
	memcpy(y, &yy, 4*sizeof(sys_float))
#endif

#ifdef __AVX512F__
#define V1FUN_PS_8(FUN, V) \
	v16sf xx = _mm512_set_ps( V, V, V, V, V, V, V, V, x7, x6, x5, x4, x3, x2, x1, x0 ); \
	v16sf yy = v1##FUN##f(xx); \
	memcpy(y, &yy, 8*sizeof(sys_float))
#else
#define V1FUN_PS_8(FUN, V) \
	v8sf xx = _mm256_set_ps( x7, x6, x5, x4, x3, x2, x1, x0 ); \
	v8sf yy = v1##FUN##f(xx); \
	memcpy(y, &yy, 8*sizeof(sys_float))
#endif

#ifdef __AVX512F__
#define V1FUN_PS_16(FUN, V) \
	v16sf xx = _mm512_set_ps( x15, x14, x13, x12, x11, x10, x9, x8, x7, x6, x5, x4, x3, x2, x1, x0 ); \
	v16sf yy = v1##FUN##f(xx); \
	memcpy(y, &yy, 16*sizeof(sys_float))
#else
#define V1FUN_PS_16(FUN, V) \
	v8sf yy[2]; \
	v8sf xx = _mm256_set_ps( x7, x6, x5, x4, x3, x2, x1, x0 ); \
	yy[0] = v1##FUN##f(xx); \
	xx = _mm256_set_ps( x15, x14, x13, x12, x11, x10, x9, x8 ); \
	yy[1] = v1##FUN##f(xx); \
	memcpy(y, &yy, 16*sizeof(sys_float))
#endif

#else // __AVX__

// fallback for non-AVX capable hardware
template<class T, class V>
inline void vecfun(const T x[], T y[], long nlo, long nhi, T (*scalfun1)(T), V (*)(V))
{
	for( long i=nlo; i < nhi; ++i )
		y[i] = scalfun1(x[i]);
}

#define V1FUN_4(FUN, V) \
	y[0] = FUN(x0); \
	y[1] = FUN(x1); \
	y[2] = FUN(x2); \
	y[3] = FUN(x3)

#define V1FUN_8(FUN, V) \
	y[0] = FUN(x0); \
	y[1] = FUN(x1); \
	y[2] = FUN(x2); \
	y[3] = FUN(x3); \
	y[4] = FUN(x4); \
	y[5] = FUN(x5); \
	y[6] = FUN(x6); \
	y[7] = FUN(x7)

#define V1FUN_16(FUN, V) \
	y[0] = FUN(x0); \
	y[1] = FUN(x1); \
	y[2] = FUN(x2); \
	y[3] = FUN(x3); \
	y[4] = FUN(x4); \
	y[5] = FUN(x5); \
	y[6] = FUN(x6); \
	y[7] = FUN(x7); \
	y[8] = FUN(x8); \
	y[9] = FUN(x9); \
	y[10] = FUN(x10); \
	y[11] = FUN(x11); \
	y[12] = FUN(x12); \
	y[13] = FUN(x13); \
	y[14] = FUN(x14); \
	y[15] = FUN(x15)

#define V1FUN_PD_4(FUN, V) \
	V1FUN_4(FUN, V)

#define V1FUN_PD_8(FUN, V) \
	V1FUN_8(FUN, V)

#define V1FUN_PS_4(FUN, V) \
	V1FUN_4(FUN##f, V)

#define V1FUN_PS_8(FUN, V) \
	V1FUN_8(FUN##f, V)

#define V1FUN_PS_16(FUN, V) \
	V1FUN_16(FUN##f, V)

#endif // __AVX__

#endif
